unit solitaire;

{
    game data structures , procedures

    moveNr = 1..31  sequence number of a move
    move   = 1..76  all possible moves
    pinlist[move].p1, .p2, .p3 translates move to pins
      (moving = delete p1,p2  add p3)

    permutation filter:
    many solutions are similar: only the sequence of moves are different
    the filter suppresses many of these solutions
    mechanism:
      each position on the board has a stack  (PMstacks[1..49])
      if a move changes a position (add or delete pin),
         the moveNr (move number  1,2,3....) is added to the stack
      for a trial move (xmove) the latest moveNr is found that changed
        the positions of the xmove
      if some moves were done between this latest moveNr and the number
      of xmove, xmove is compared to all in-between moves
      xmove is skipped if any in-between move is larger then xmove
        (because this move combination occurred before already)
}

interface

uses forms,classes,controls,types,sysutils,graphics,printers,dialogs{};

type TmainBtn = (mbSearch,mbReplay,mbPlay,mbPlace,mbOff);
     TmoveDirection = (mdForward,mdBackward);
     TMouseAction = (maDown,maMove,maUp);
     type TSearchState = (ssTimeout,ssSolved,ssEnd);

procedure drawPin(nr : byte);
procedure erasePin(nr : byte);
procedure resetboard;
procedure showmove(nr : byte);
procedure showallMoves;
procedure showboard;
function solve : TsearchState;
procedure playStepBackward;
procedure replaystepforward;
procedure replaystepbackward;
procedure replaytimer;
procedure drawMovePointer(nr : byte);
procedure repaintbox2;
procedure initgame;
procedure initreplaymode;
procedure exitReplaymode;
procedure initPlayMode;
procedure initsearchmode;
procedure initplacemode;
procedure mouseProc;
procedure placeProc;
procedure clearPrintlist;
procedure addToPrintlist;
procedure printsolutions;
procedure makemovePinList;
procedure clearPMstacks;
procedure cleargame;
procedure setstartboard(row,im : byte);
procedure savegame;
procedure loadgame;
procedure startfromfile;

var mainbutton : TmainBtn;
    mouseAction : TMouseAction;
    mouseX,mouseY : integer;
    solnr : integer;                       //number of solution

implementation

uses unit1;

procedure clearmoveDisplay(nr : byte);forward;

type Tmove123 = record             //new definition of move
                  p1,p2,p3 : byte;  //p1,p2 removed, p3 added
                 end;

     TmovePath = record
                  valid : boolean;        //if valid definition
                  pin3  : byte;           //destination on board
                  x1,y1,x2,y2 : integer;
                 end;

     Tprintdata = record
                   bottomNr : byte;       //below is printed green
                   topNr    : byte;       //lst move in list
                  end;

     TsaveFormat = record
                    startboard : array[1..49] of byte;
                    moves : array[1..31] of byte;
                    bottomMoveNr : byte;
                    topMoveNr : byte;
                   end;

const movedirs : array[1..49] of byte =    //1: right
                 (0, 0, 9, 8,12, 0, 0,     //2: up
                  0, 0, 9, 8,12, 0, 0,     //4: left
                  9, 9,15,15,15,12,12,     //8: down
                  1, 1,15,15,15, 4, 4,
                  3, 3,15,15,15, 6, 6,
                  0, 0, 3, 2, 6, 0, 0,
                  0, 0, 3, 2, 6, 0, 0);

      pintype  : array[1..49] of byte =
                 (0, 0, 3, 2, 3, 0, 0,     //1: win
                  0, 0, 2, 1, 2, 0, 0,     //2: win enable
                  3, 2, 3, 2, 3, 2, 3,     //3: corner
                  2, 1, 2, 1, 2, 1, 2,
                  3, 2, 3, 2, 3, 2, 3,
                  0, 0, 2, 1, 2, 0, 0,
                  0, 0, 3, 2, 3, 0, 0);

      keys     : array[1..49] of byte =    //xor of all pins must be 2 =
                 (0, 0, 1, 2, 3, 0, 0,     //middle value
                  0, 0, 2, 3, 1, 0, 0,
                  1, 2, 3, 1, 2, 3, 1,
                  2, 3, 1, 2, 3, 1, 2,
                  3, 1, 2, 3, 1, 2, 3,
                  0, 0, 3, 1, 2, 0, 0,
                  0, 0, 1, 2, 3, 0, 0);

      maxprintlist = 20;                  //max nr of solutions in print list

var PinList : array[1..76] of Tmove123;   //definition of moves
    board : array[1..49] of byte;         //0: empty   1: pin
    startboard : array[1..49] of byte;    //initial board state (selectable)
    moves : array[1..31] of byte;         //list of moves
    moveNr : byte;                        //next move : 1..32 (end of list)
    oldmovePtr : byte;                    //old arrow position in move table
    BottomMoveNr : byte;                  //1st changeable move nr in list
    TopMoveNr : byte;                     //highest move nr in movelist
    movelimit : byte = 32;                //check for solution check at this nr
    movedirection : Tmovedirection;       //forward,backward
    p1count,p2count,p3count : byte;       //count of pin types 1,2,3
    rpmotion : boolean;

    PMStacks : array[1..49,0..30] of byte; //49 stacks of moves
    PMHeight : array[1..49] of byte;       //entries per stack

    timeout : longInt;              //search interrupt timer
    rpdx,rpdy  : shortInt;          //replay dx,dy movement per timer step
    rpRect     : Trect;             //       ball rectangle
    rp1,rp2,rp3: byte;              //       balls
    rpbusy     : boolean = false;   //timer is busy with move
    rpcount    : byte;              //timer actions control counter
    mcount : byte;                  //mouseAction sequence counter
    onPin1 : boolean;               //mousepointer over pin
    offsetX,offsetY : shortInt;     //pinXY - mousedownXY
    playPin : byte;
    destpin : byte;                 //destination position on board
    playRect : Trect;               //rectangle of moving pin
    movepaths : array[1..4] of Tmovepath;

    printlist : array[1..maxprintlist,1..31] of byte;//[x,0] : solNr (solution)
    printdata : array[1..maxprintlist] of Tprintdata;
    topPrintlist : byte;
    prSolNrs  : array[1..maxprintlist] of longInt;//nr of solution

    savedata : Tsaveformat;
    savefile : file of Tsaveformat;

// --- housekeeping -------

procedure setstartboard(row,im : byte);
//set row of startboard[..] to image im
//called when new selection is made
var i,k : byte;
begin
 k := $80;
 for i := 7*row-6 to 7*row do
  begin
   k := k shr 1;
   if (im and k) = k then startboard[i] := 1
    else startboard[i] :=0;
   end;//for
end;

function GetPincount : byte;
//count the number of pins on the board
var i : byte;
begin
 result := 0;
 for i := 1 to 49 do if board[i] = 1 then inc(result);
end;

procedure adjustPinTypeCount(pin : byte; v : shortInt);
//add v to count of pintype
begin
 case pintype[pin] of          //get type of pin (1,2,3)
  1 : p1count := p1count + v;
  2 : p2count := p2count + v;
  3 : p3count := p3count + v;
 end;//case
end;

procedure setP123counts;
//scan board[ ] to adjust pintype counts
var i : byte;
begin
 p1count := 0;
 p2count := 0;
 p3count := 0;
 for i := 1 to 49 do
  if board[i] = 1 then
   case pintype[i] of
    1 : inc(p1count);
    2 : inc(p2count);
    3 : inc(p3count);
   end;//case
end;  

procedure makeMovePinList;
//moves 1..76 translated to pin nrs in PinList[ ]
const u : array[0..3] of shortint = (1,-7,-1,7);  //pin direction increments
var m,n,d : byte;                                   //m:movepin1  d:direction
begin
 n := 0;
 for m := 3 to 47 do
  for d := 0 to 3 do
   if movedirs[m] and (1 shl d) <> 0 then
    begin
     inc(n);                //next entry in table
     with PinList[n] do
      begin
       p1 := m;
       p2 := p1 + u[d];
       p3 := p2 + u[d];
      end;
   end;
end;

function dir2arrow(d : byte) : char;
//direction code 0..3 to arrow for move display
begin
 case d of
  0 : result := #174;
  1 : result := #173;
  2 : result := #172;
  3 : result := #175;
 end;
end;

//---game presetting ---

procedure resetboard;
//set board to initial conditions
//point to 1st move
//do not erase movelist
var i,j : byte;
begin
 for i := 1 to 49 do
  board[i] := startboard[i];
 showboard;
 moveNr := 1;
 setP123counts;
 movelimit := GetPincount;
 clearPMstacks;
end;

procedure cleargame;
var i : byte;
begin
 for i := 1 to 49 do
  begin
   board[i] := 0;
   startboard[i] := 0;
   PMheight[i] := 0;
  end;
 showboard;
 setP123counts;
 SolNr := 0;
 bottommoveNr := 1;
 TopMoveNr := 0;
 moveNr := 1;
 form1.sollabel.caption := '';
end;

procedure initgame;
//erase movelist, reset board
var i : word;
begin
 for i := 1 to 31 do moves[i] := 0;
 for i := 1 to 32 do clearmoveDisplay(i);
 solnr := 0;
 resetboard;
 bottommoveNr := 1;
 topMoveNr := 0;
 form1.sollabel.caption := '';
end;

procedure initPlayMode;
//when menu play button pressed
begin
 bottomMoveNr := moveNr;
 showallmoves;
 mcount := 0;
 onPin1 := false;
 drawmovepointer(moveNr);
end;

procedure initReplaymode;
//if replaybutton pressed
begin
 resetboard;
 drawMovePointer(moveNr);
end;

procedure exitReplaymode;
//called if exiting replaymode
//moveNr may be < TopMove
//truncate movelist to
var i : byte;
begin
 TopMoveNr := moveNr - 1;
 if bottomMoveNr > moveNr then bottommoveNr := movenr;
 for i := moveNr to 31 do moves[i] := 0;  //remove unused moves
 showAllMoves;
end;

procedure initsearchmode;
//called when search mode is selected
//test pin keys for "no solution"
//set movelimit (= last possible move)
var i : byte;
    xrsum : byte;
begin
 moveNr := topmoveNr+1;
 drawmovepointer(moveNr);
 movelimit := GetPincount + topMoveNr;
 clearprintlist;
 xrsum := 2;
 for i := 1 to 49 do
  if board[i] = 1 then xrsum := xrsum xor keys[i];
 if xrsum <> 0 then form1.msgbox.Caption := 'warning: no solution';
end;

procedure initplacemode;
//clear moves table
//copy board ---> startboard
var i : byte;
begin
 for i := 1 to 49 do startboard[i] := board[i];
 initgame;
 showmove(moveNr);
end;

//--- P system --- (eliminate same board positions, skip permutations)

procedure clearPMstacks;
var i,j : byte;
begin
 for i := 1 to 49 do   //Permutation filter system
  begin
   PMheight[i] := 0;
   for j := 0 to 30 do PMstacks[i,j] := 0;
  end;
end;

procedure PMpush(m,nr : byte);
//push nr to stack of board position pins of move m
var pin,h,i : byte;
begin
 for i := 1 to 3 do
  begin
   case i of
    1 :  pin := pinlist[m].p1;  //get pin1 of move m
    2 :  pin := pinlist[m].p2;  //       2
    3 :  pin := pinlist[m].p3;  //       3
   end;
   h := PMheight[pin] + 1;
   PMstacks[pin,h] := nr;
   PMheight[pin] := h;
  end;
end;

procedure PMpop(m : byte);
//remove last entry from stack for pins of move m
var pin,x,i : byte;
begin
 for i := 1 to 3 do
  begin
   case i of
    1 :  pin := pinlist[m].p1; //get pin1 of move m
    2 :  pin := pinlist[m].p2; //       2
    3 :  pin := pinlist[m].p3; //       3
   end;
   dec(PMheight[pin]);
  end;
end;

function lastPmove(m : byte) : byte;
//find latest moveNr that changed pin positions of move m
var pin,x,i : byte;
begin
 result := 0;
 for i := 1 to 3 do
  begin
   case i of
    1 :  pin := pinlist[m].p1;
    2 :  pin := pinlist[m].p2;
    3 :  pin := pinlist[m].p3;
   end;
   x := PMstacks[pin,PMheight[pin]];  //x=latest moveNr modifying p1,p2,p3
   if x <> 0 then
    if x > result then result := x;
  end;
end;

// --- painting ---

function pinToRect(nr : byte) : Trect;
begin
 with result do
  begin
   left := ((nr-1) mod 7) * 40;
   top := ((nr-1) div 7) * 40;
   right := left + 40;
   bottom := top + 40;
  end;
end;

procedure repaintbox2;
//on paint of box2
var i : byte;
    x,y : integer;
    s : string;
begin
 with form1.paintbox2 do with canvas do
  begin
   brush.Color := $f0ffff;
   fillrect(rect(0,0,width-1,height-1));
   font.Name := 'arial';
   font.Color := $808080;
   font.Height := 18;
   for i := 1 to 31 do
    begin
     s := inttostr(i)+':';
     x := 25 - textwidth(s);
     y := (i-1)*20;
     textout(x,y,s);
    end;
  end;
 showAllMoves;
 drawMovepointer(moveNr);
end;

procedure drawPin(nr : byte);
//draw in map1
var r : Trect;
begin
 if movedirs[nr] = 0 then exit; //no pin here

 r := pinToRect(nr);
 map1.canvas.draw(r.left,r.top,form1.ball.picture.bitmap);
 updatemap1(r);
end;

procedure erasePin(nr : byte);
//erase pin nr in map1
var r : Trect;
begin
 if movedirs[nr] = 0 then exit; //no pin here

 r := pinToRect(nr);
 updatemap1(r);
 inc(r.left,2);
 inc(r.top,2);
 dec(r.right,2);
 dec(r.bottom,2);
 with map1 do with canvas do
  begin
   brush.color := fieldBGcolor;
   fillrect(r);
   inc(r.left,18);
   inc(r.top,18);
   pen.color := $c0c0c0;
   brush.color := $ffffff;
   pen.width := 2;
   ellipse(r.left-10,r.top-10,r.left+10,r.top+10);
  end;
end;

//------- board --- drawing --- support ---

procedure makeP23(var p1,p2,p3:byte; dir : byte);
//p1 : pin Nr of move, dir: direction
//given p1,dir ....  make p2,p3 pins of move
begin
 case dir of
  1 : begin
       p2 := p1+1; p3 := p1+2;
      end;
  2 : begin
       p2 := p1-7; p3 := p1-14;
      end;
  4 : begin
       p2 := p1-1; p3 := p1-2;
      end;
  8 : begin
       p2 := p1+7; p3 := p1+14;
      end;
 end;
end;

function XYtoField(x,y : integer; var pin : byte) : boolean;
//return true if (x,y) on valid field
//if true , return pin (field) nr of (x,y)
var modx,mody : byte;
begin
 result := false;
 modx := x mod 40;
 mody := y mod 40;
 if (modX > 5) and (modX < 35) and (modY > 5) and (modY < 35) then
  begin
   pin := (x div 40) + (y div 40)*7 + 1;
   if movedirs[pin] <> 0 then result := true;
  end;
end;

function XYtoPin1(x,y:integer; var pin : byte) : boolean;
//translate x,y for box1 to pin position
//true if move OK with this pin
var p1,p2,p3,dir : byte;
begin
 result := false;
 dir := 1;
 if XYtoField(x,y,p1) then
  repeat
   if (dir and movedirs[p1]) > 0 then
    begin
     makeP23(p1,p2,p3,dir);
     result := (board[p1] = 1) and (board[p2] = 1) and (board[p3] = 0);
    end;
   dir := dir shl 1;
  until result or (dir = 16);
 if result then pin := p1;
end;

procedure makemovepaths;
//setup movepath for playpin
//direction counts 1,2,4,8
var i,dir,p2,p3 :byte;
    r1,r2 : Trect;
begin
 for i := 1 to 4 do movepaths[i].valid := false;//set all path invalid
 dir := 1;
 for i := 1 to 4 do
  begin
   if (movedirs[playpin] and dir) > 0 then
    begin
     makeP23(playpin,p2,p3,dir);
     if (board[p2] = 1) and (board[p3] = 0) then
      begin
       r1 := pintorect(playpin);
       r2 := pintorect(p3);
       with movepaths[i] do
        begin
         valid := true;
         pin3 := p3;
         case dir of
          1,8 : begin
                 x1 := r1.left-15;
                 y1 := r1.top-15;
                 x2 := r2.left+15;
                 y2 := r2.top+15;
                end;
          2,4 : begin
                 x1 := r2.left-15;
                 y1 := r2.top-15;
                 x2 := r1.left+15;
                 y2 := r1.top+15;
                end;
         end;//case
        end;//with
      end;//if
    end;//if
   dir := dir shl 1; //next direction
  end;//for
end;

function onpath(x,y : integer; var destpin : byte) : boolean;
//dir : direction of move
//return true if ball moves along correct path
var i : byte;
begin
 result := false;
 i := 0;      //path nr
 repeat
  inc(i);
  with movepaths[i] do
   begin
    if valid then result := (x > x1) and (x < x2) and (y > y1) and (y < y2);
    if result then destpin := pin3;
   end;//with
 until result or (i = 4);
end;

function move2dir(m : byte) : byte;
//translate move 1..76 to direction
var x1,y1,x2,y2 : byte; //pin position of column,row
    p1,p3 : byte;
begin
 p1 := pinlist[m].p1;
 p3 := pinlist[m].p3;
 x1 := (p1-1) mod 7;
 x2 := (p3-1) mod 7;
 y1 := (p1-1) div 7;
 y2 := (p3-1) div 7;
 if x1 = x2 then            //vertical motion
  begin
   if y1 > y2 then result := 1 else result := 3;
  end
 else                       //horizontal motion
  if x1 < x2 then result := 0 else result := 2;
end;     

procedure setRPdXY;
//set rpdx,rpdy acc. to movedirection
//is XY step for pin movement
var dir : byte;
    f : shortint;
begin
 dir := move2dir(moves[moveNr]);
 if movedirection = mdForward then f := 1 else f := -1;
  case dir of
    0 : begin
         rpdx := 2*f; rpdy := 0;
        end;
    1 : begin
         rpdx := 0; rpdy := -2*f;
        end;
    2 : begin
         rpdx := -2*f; rpdy := 0;
        end;
    3 : begin
         rpdx := 0; rpdy := 2*f;
        end;
   end;//case
end;

procedure showBoard;
var i : byte;
//draw balls according to array board
begin
 for i := 1 to 49 do
  if board[i] = 1 then drawpin(i)
   else erasePin(i);
 copymap12;
 copymap2box;
end;

//--------- search for solutions ------------

function solve : TsearchState;
//solve solitaire game
//speedup search by pin differentiation
//xmove is trial move during search
//moveNr points to next free entry in moves table
//exit: ssTimeout, ssSolved, ssEnd
var m,pm,p1,p2,p3: byte;
    xmove : byte;
    xp1count,xp2count,xp3count : byte;
label start,testmove,nextmove,moveback;
begin

 timeout := 100000;      //interrupt timer count
 stopflag := false;

start :                  //(re)start with 1st move

 dec(timeout);
 if timeout = 0 then
  begin
   application.ProcessMessages;
   if stopflag then              //test stop btn was pressed
    begin
     result := ssTimeOut;
     topmoveNr := moveNr - 1;
     exit;
    end
    else timeout := 100000;
  end;

 xmove := 1;             //1st trial move

testmove :               //test move is possible

 p1 := pinlist[xmove].p1;
 if board[p1] = 0 then goto nextmove;                      //no pin
 p2 := pinlist[xmove].p2;
 if board[p2] = 0 then goto nextmove;                      //no pin
 p3 := pinlist[xmove].p3;
 if board[p3] = 1 then goto nextmove;                      //no hole

//------- P filter -------------

 if form1.Pfilter1.Checked then
  begin
   pm := lastPmove(xmove) + 1;
   while pm < moveNr do
    begin
     if moves[pm] > xmove then goto nextmove;
     inc(pm);
    end;
  end;  

//------- pin differentiation ---

 xp1count := p1count;
 xp2count := p2count;
 xp3count := p3count;

 case pintype[p2] of
  1 : dec(xp1count);
  2 : dec(xp2count);
  3 : dec(xp3count);
 end;//case

 if (xp1count = 0) or (xp2count < xp3count) then goto nextmove;

//---- record move -----

 board[p1] := 0;
 board[p2] := 0;
 board[p3] := 1;
 moves[moveNr] := Xmove;
 PMpush(xmove,moveNr);
 p1count := xp1count;
 p2count := xp2count;
 p3count := xp3count;

//--- test solution ----

 inc(moveNr);
 if (moveNr = movelimit) and (board[25] = 1) then
  begin
   result := ssSolved;
   topmoveNr := moveNr-1;
   exit;
  end
  else goto start;

nextmove :

 inc(xmove);
 if xmove < 77 then goto testmove;  //move range [1..76]

moveback :

 if moveNr <= BottomMoveNr then
  begin
   result := ssEnd;
   topmoveNr := movenr - 1;
   exit;
  end;

 dec(moveNr);                     //point to previous move
 m := moves[moveNr];
 p1 := pinlist[m].p1;
 p2 := pinlist[m].p2;
 p3 := pinlist[m].p3;
 board[p1] := 1;
 board[p2] := 1;
 board[p3] := 0;
 PMpop(m);
 xmove := moves[moveNr];

 case pintype[p2] of
  1 : inc(p1count);
  2 : inc(p2count);
  3 : inc(p3count);
 end;

 goto nextmove;

end;

//------------ replay -------------

procedure replaytimer;
//called by timer
//slowly move ball
//rpcount is timer sequence counter
//rpmotion enables motion of ball
//move is 2 pixels per count while rpmotion is true
begin
 if movedirection = mdForward then
  case rpcount of
    0 : begin
         erasepin(rp1);
         rpmotion := false;
        end;
   20 : begin
         drawpin(rp1);
         rpmotion:= false;
        end;
   30 : begin
         erasepin(rp1);
         rpmotion := true;
        end;
   50 : erasepin(rp2);
   70 : begin
         drawpin(rp3);
         form1.Timer1.Enabled := false;
         rpBusy := false;
        end;
  end;//case
 if movedirection = mdbackward then
 case rpcount of
    0 : begin
         erasepin(rp3);
         rpmotion := false;
        end;
   20 : begin
         drawpin(rp3);
         rpmotion := false;
        end;
   30 : begin
         erasepin(rp3);
         rpmotion := true;
        end;
   50 : drawpin(rp2);
   70 : begin
         drawpin(rp1);
         form1.Timer1.Enabled := false;
         rpBusy := false;
        end;
  end;//case
 copymap12;
 if rpmotion then
  begin
   map2.Canvas.Draw(rpRect.Left,rpRect.Top,form1.ball.Picture.bitmap);
   addrects(rpRect);
  end;
 copymap2box;

 inc(rpcount);
 if rpmotion then
  with rpRect do
   begin
    left := left + rpdx;
    right := right + rpdx;
    top := top + rpdy;
    bottom := bottom + rpdy;
   end;
end;

procedure replayStepForward;
//called if 'step' btn clicked in REPLAY mode
var m : byte;
begin
 if (moveNr > topMoveNr) or rpbusy then exit;

 movedirection := mdForward;
 m := moves[moveNr];
 rp1 := pinlist[m].p1;
 rp2 := pinlist[m].p2;
 rp3 := pinlist[m].p3;
 board[rp1] := 0;
 board[rp2] := 0;
 board[rp3] := 1;
 adjustpintypeCount(rp2,-1);
 PMpush(m,moveNr);
 rpbusy := true;
 rprect := pinToRect(rp1);
 setRPdXY;
 rpcount := 0;
 form1.timer1.enabled := true;
 form1.msgbox.Caption := 'move '+inttostr(moveNr);
 if moveNr <= 31 then inc(moveNr);
 drawmovepointer(moveNr);//point to next move
end;

procedure replayStepBackward;
//called if BACK btn clicked in REPLAY mode
var m : byte;
begin
 if (moveNr <=1) or rpbusy then exit;

 movedirection := mdBackward;
 dec(moveNr);
 m := moves[moveNr];
 rp1 := pinlist[m].p1;
 rp2 := pinlist[m].p2;
 rp3 := pinlist[m].p3;
 board[rp1] := 1;
 board[rp2] := 1;
 board[rp3] := 0;
 PMpop(m);
 adjustpintypeCount(rp2,1);
 rpbusy := true;
 rprect := pinToRect(rp3);
 setRPdXY;
 rpcount := 0;
 form1.timer1.enabled := true;
 form1.msgbox.Caption := '';
 drawmovepointer(moveNr);//point to next move
end;

//------- player moves ---------------------

procedure registerMove(p1,p3 : byte);
//player move from p1 to p3
var i,p2 : byte;
    hit : boolean;
begin
 board[p1] :=0;
 p2 := (p1 + p3) div 2;
 board[p2] :=0;
 board[p3] :=1;
 adjustPinTypeCount(p2, -1);
 hit := false;
 i := 1;
 repeat
  if (pinlist[i].p1 = p1) and (pinlist[i].p3 = p3) then hit := true
   else inc(i);
 until hit;
 moves[moveNr] := i;
 PMpush(i,moveNr);
 TopmoveNr := moveNr;
 showmove(moveNr);
 inc(moveNr);
 showmove(moveNr);
 drawmovepointer(moveNr);
end;

procedure playStepBackward;
begin
 if rpbusy then exit;

 clearmovedisplay(moveNr);
 replaystepbackward;
 bottomMoveNr := topMoveNr;
 if topmoveNr > 0 then dec(topMoveNr);
 showAllmoves;
 drawMovePointer(moveNr);
end;

//--- all mouse events of paintbox1 ----

procedure mouseproc;
//called by all mouse events, params mouseX,mousY,mouseAction
//mcount :0  moving over board
//       :1  mousedown over ball
//       :2  mouseMove with ball
//all mouse events in case of play
var onPin2 : boolean;
    pin : byte;
begin
 with map2.canvas do
 case mouseaction of
  maDown :
        case mcount of
         0 : if onPin1 then
              begin
               mcount := 1;
               playrect := pinToRect(playPin);
               offsetX := playrect.left - mouseX;
               offsetY := playrect.Top - mouseY;
              end; 
         1 :;
         2 :;
        end;
  maMove :
        case mcount of
         0 : begin
              onPin2 := XYtoPin1(mouseX,mouseY,playpin);
              if onPin2 <> onPin1 then
               if onPin2 then form1.PaintBox1.cursor := crHandpoint
                else form1.PaintBox1.cursor := crArrow;
              onPin1 := onPin2;
             end;
         1 : begin
              makemovePaths;       //for playpin
              erasePin(playPin);   //remove from map1
              mcount := 2;
             end;
         2 : begin
              mouseX := mouseX + offsetX;
              mouseY := mouseY + offsetY;
              onPin1 := onpath(mouseX,mouseY,destpin);
              if onPin1 then
               begin
                copymap12;
                with playrect do
                 begin
                  left := mouseX; right := left + 40;
                  top := mouseY; bottom := top + 40;
                 end;
                draw(mouseX,mouseY,form1.ball.picture.bitmap);
                addrects(playrect);
                copymap2box;
               end;
             end;
        end;//case
  maUp   :
        begin
         if mcount = 2 then
          begin
           onPin1 := onPin1 and XYtoField(mouseX+20,mouseY+20,pin);
           onPin1 := onPin1 and (pin = destpin);
           if onPin1 then
            begin
             erasePin((playPin + destpin) div 2);
             drawPin(destpin);
             bottomMoveNr := moveNr+1;
             registerMove(playPin,destPin);
             if (moveNr = movelimit) and (board[25] = 1) then
              form1.msgbox.caption := 'solution found, congratulations!'
              else form1.msgbox.caption := 'OK';
            end
           else
            begin
             drawPin(playpin);
             form1.msgbox.Caption := 'move rejected';
            end;
           copymap12;
           copymap2box;
          end;//if
         form1.PaintBox1.cursor := crArrow;
         mcount := 0;
        end;
 end;//case
end;

procedure Placeproc;
//add , remove pin at mouseX,mousey coordinates
var pin : byte;
    d : smallint;
begin
 if XYtofield(mouseX,mouseY,pin) then
  begin
   board[pin] := board[pin] xor 1;
   startboard[pin] := board[pin];
   if board[pin] = 1 then
    begin
     drawpin(pin);
     d := 1;
    end
     else begin
           erasePin(pin);
           d := -1;
          end;
   copymap12;
   copymap2box;
   setP123counts;
  end;//if
end;

//--- move list ---

procedure drawMovePointer(nr : byte);
//paint arrow in paintbox2 pointing to next move
begin
 with form1 do with paintbox2.Canvas do
  begin
   if oldmovePtr > 0 then
    begin
     brush.style := bsSolid;
     brush.Color := $f0ffff;
     fillrect(rect(40,(oldmovePtr-1)*20,60,oldmovePtr*20));
     oldmovePtr := 0;
    end;
   if (nr > 0) and (nr <= topmoveNr+1) then
    begin
     draw(40,(nr-1)*20,arrowimage.picture.bitmap);
     oldmovePtr := nr;
    end;
 end;
end;

procedure clearmovedisplay(nr : byte);
//erase move nr in paintbox 2 moves list
var r :Trect;
begin
 r.Top := (nr-1)*20;
 r.Bottom := (r.Top+20);
 r.Left := 30;
 r.Right := form1.paintbox2.width;
 with form1.PaintBox2.Canvas do
  begin
   brush.color := fieldBGcolor;
   brush.style := bssolid;
   fillrect(r);
  end;
end;

procedure showmove(nr : byte);
//show move nr in paintbox2  "moves" display
//if nr < bottomMoveNr ---> draw green number
//if nr = topmove+1 ---> display <end>
var x,y : integer;
    s : string;
    m,dir : byte;
begin
 clearmoveDisplay(nr);
 with form1.PaintBox2.canvas do
  begin
   y := (nr-1)*20;
   brush.Style := bsClear;
   if nr <= topmoveNr then
    begin
     font.Name := 'arial';
     if nr < bottomMoveNr then font.color := $00a000 else font.Color := $0;
     m := moves[nr];
     dir := move2dir(m);
     s := inttostr(pinlist[m].p1);
     x := 80-textwidth(s);
     textout(x,y,s);
     s := dir2arrow(dir);
     x := 100-textwidth(s);
     font.Name := 'symbol';
     textout(x,y,s);
    end;
   if nr = topmoveNr+1 then
    begin
     s := '< end >';
     x := 60;
     font.Name := 'arial';
     font.Color := 0;
     textout(x,y,s);
    end;
  end;//with
end;

procedure showallMoves;
//list all moves in paintbox2
var i : byte;
begin
 for i := 1 to 32 do showmove(i);
end;

//----------- printer --------------------

procedure clearPrintlist;
var i,j : byte;
begin
 topprintlist := 0;
 for i := 1 to maxprintlist do
  begin
   prsolNrs[i] := 0;
   printdata[i].bottomNr := 0;
   printdata[i].topNr := 0;
   for j := 1 to 31 do printlist[i,j] := 0;
  end;//for
 form1.prlist.Caption := '0';
 form1.printmsg.Caption := '';
end;

procedure AddToPrintList;
var i : byte;
begin
 if topprintlist < maxprintlist then
  begin
   inc(topprintlist);
   if moveNr < movelimit then
    begin
     form1.printmsg.Caption := 'incomplete solution';
     exit;
    end;
    with printdata[topprintlist] do
     begin
      bottomNr := bottomMoveNr;
      topNr := topMoveNr;
     end;
    for i := 1 to topMoveNr do printlist[topprintlist,i] := moves[i];
    printdata[topprintlist].bottomnr := bottomMoveNr;
    prSolNrs[topprintlist] := solNr;
    form1.printmsg.Caption := 'solution '+ inttostr(solNr)+
                              ' added to print list';
   form1.prList.caption := inttostr(topprintlist);
  end
   else form1.printmsg.Caption := 'print queue full';
end;

procedure printsolutions;
//list 1..20 solutions
var i,j,dv,dh,ytop,pin,dir : byte;
xleft : word;
    s : string;
    arrow : char;
    x1,x2,y1,y2 : integer;
    mult : byte;
    bm : Tbitmap;
begin
 form1.printmsg.Caption := 'printing...';
 form1.PrintScale := poNone;
 with printer do with canvas do
  begin
   orientation := poLandscape;
   mult := pageWidth div 2400;//1, 2
   dv := 70*mult; dh := 125*mult;    //hor,vert dimension of field
   xleft := 750*mult;
   ytop := trunc(1.25*dv);

   begindoc;

   pen.Width := 3*mult;
   brush.Style := bsClear;
   x1 := xleft; x2 := xleft + 21*dh;             //horizontal fat lines
   y1 := ytop;  moveto(x1,y1); lineto(x2,y1);
   y1 := ytop+dv;  moveto(x1,y1); lineto(x2,y1);
   y1 := ytop+32*dv; moveto(x1,y1); lineto(x2,y1);
   font.Name := 'arial';
   font.Height := trunc(1.2*dv);
   font.Style := [fsBold];
   font.color := $000000;
   y1 := 0;
   textout(xleft,y1,'solitaire solutions');         //header

   font.Style := [];
   y1 := ytop; y2 := ytop+32*dv;
   for i := 1 to 22 do                          //vertical fat lines
    begin
     x1 := xleft + dh*(i-1);
     moveto(x1,y1); lineto(x1,y2);
    end;

   moveto(xleft,ytop); lineto(xleft+dh,ytop+dv);
   font.Height := 30*mult;
   textout(xleft+trunc(0.4*dh),ytop,'sol#');
   textout(xleft+trunc(0.1*dh),trunc(ytop + 0.6*dv),'move');

   pen.Width := mult;
   pen.Color := $808080;
   for i := 1 to 30 do
    begin
     y2 := ytop+(i+1)*dv;
     moveto(xleft,y2); lineto(xleft+21*dh,y2);   //horizontal small lines
    end;

   font.Height := trunc(0.33*dh);
   for i := 1 to 31 do
    begin
     s := inttostr(i);
     x1 := xleft + trunc(0.75*dh) -textwidth(s);
     y1 := trunc(ytop+(0.2+i)*dv);
     textout(x1,y1,s);//moves 1..31            //moves 1..31
    end;

   for i := 1 to topPrintlist do               //for all games
    begin
     font.Name := 'arial';
     s := inttostr(prSolNrs[i]);
     y1 := trunc(ytop + 0.2*dv);
     x1 := trunc(xleft + (i+0.75)*dh - textwidth(s)); //solution number
     textout(x1,y1,s);
     x1 := xleft + i*dh;
     for j := 1 to printdata[i].topNr do
      begin                                   //for all moves
       font.Name := 'arial';
       if j < printdata[i].bottomNr then font.Color := $00a000
        else font.color := $000000;
       y1 := trunc(ytop + (j+0.2)*dv);
       pin := pinlist[printlist[i,j]].p1;
       s  := inttostr(pin);
       x2 := x1 + trunc(0.6*dh) - textwidth(s);
       textout(x2,y1,s);                      //move
       dir := move2dir(printlist[i,j]);
       s  := dir2arrow(dir);
       font.Name := 'symbol';
       x2 := x1 + trunc(0.9*dh) - textwidth(s);
       textout(x2,y1,s);                      //direction arrow
      end;
    end;//for i

// draw game

   if mult = 1 then bm := form1.prball64.picture.bitmap
    else bm := form1.prball128.picture.bitmap;
   xleft := 25*mult;
   pen.Width := 6*mult;
   pen.Color := $000000;
   roundrect(xleft,ytop,xleft+700*mult,ytop+700*mult,32*mult,32*mult);
   pen.Width := 3*mult;
   for i := 1 to 49 do
    begin
     x1 := xleft + ((i-1) mod 7) * 100 * mult;
     y1 := ytop  + ((i-1) div 7) * 100 * mult;
     if movedirs[i] <> 0 then
      if startboard[i] = 1 then draw(x1+18*mult,y1+18*mult,bm)
       else ellipse(x1+32*mult,y1+32*mult,x1+68*mult,y1+68*mult);
    end;

   enddoc;
  end;//with printer
 form1.printmsg.caption := '';
 clearprintlist;
end;

// --- save and reload

procedure savegame;
//saveBtn click
var i : byte;
begin
 for i := 1 to 49 do savedata.startboard[i] := startboard[i];
 for i := 1 to 31 do savedata.moves[i] := moves[i];
 savedata.bottomMoveNr := bottomMoveNr;
 savedata.topMoveNr := topmoveNr;
 with form1.SaveDialog1 do
  if execute then
   try
    try
     assignfile(savefile,filename);
     rewrite(savefile);
     write(savefile,savedata);
    except
     form1.msgbox.Caption := 'cannot write file';
    end
   finally
    closefile(savefile);
    form1.saveBtn.Down := false;
   end;
end;

procedure rebuiltgame;
var i : byte;
begin
 for i := 1 to 49 do startboard[i] := savedata.startboard[i];
 for i := 1 to 31 do moves[i] := savedata.moves[i];
 bottomMoveNr := savedata.bottomMoveNr;
 topMoveNr := savedata.topmoveNr;
 showallmoves;
 drawmovepointer(moveNr);
 resetboard; //clearPMstacks, setP123counts,set movelimt, copy startboard -->
             //movePtr := 1
 mainbutton := mbReplay;
 form1.replaybtn.down := true;
 with form1 do
  begin
   gobtn.Enabled := true;
   backbtn.Enabled := true;
  end;
 setReplayCaptions;
 for i := 1 to moveNr-1 do PMpush(moves[i],i); //rebuilt PMstacks
end;

procedure loadgame;
var OK : boolean;
    i : byte;
begin
 with form1.OpenDialog1 do
  if execute then
   try
    try
     assignfile(savefile,filename);
     reset(savefile);
     read(savefile,savedata);
     OK := true;
    except
     form1.msgbox.Caption := 'cannot read file';
     OK := false;
    end
   finally
    closefile(savefile);
    form1.loadBtn.Down := false;
   end;
 if OK then rebuiltgame;
end;

procedure startfromfile;
//called from form create , if started from *.sol file
var OK : boolean;
begin
 try
  try
   assignfile(savefile,paramstr(1));
   reset(savefile);
   read(savefile,savedata);
   OK := true;
  except
   form1.msgbox.Caption := 'cannot open game';
   OK := false;
  end
 finally
  closefile(savefile);
 end;
 if OK then rebuiltgame;
end;

end.
