unit timer_unit;

interface

uses windows;

type TI64 = array[0..8] of byte;

function diff64(v1,v2 : TI64) : TI64;
function I64tofloat(I64 : TI64) : extended;
procedure GetCPUTicks(var i64 : TI64);
procedure setCPUclock;

var
 CPUClock: extended;

implementation

procedure GetCPUTicks(var i64 : TI64);
//store CPU ticks in array[0..7]
var i : byte;
label loop1,loop2;
begin
 asm
   DB $0F,$31  // RDTSC command
               // put CPU ticks in EAX (bits 0..31) and EDX (32..63)
   mov ECX,i64 //save address
   mov i,4
loop1:
   mov [ECX],AL
   shr EAX,8
   inc ECX
   dec i
   jnz loop1
   mov i,4
loop2:
   mov [ECX],DL
   shr EDX,8
   inc ECX
   dec i
   jnz loop2
 end;
end;

procedure setCPUclock;
//calculate the clock speed
var t1,t2 : TI64;
    t : DWORD;
begin
 t := GetTickCount;
 while t=GetTickCount do;
 GetCPUTicks(t1); //get the time-stamp counter value
 t := t + 500;
 while GetTickCount < t do;   // delay for 0,5 sec
 GetCPUTicks(t2);  // clock cycle number in 0,5 second
 t1 := Diff64(t1,t2);
 CPUclock := I64ToFloat(t1)*2e-6;
end;

function diff64(v1,v2 : TI64) : TI64;
//difference v2 - v1 of 2 64 bit integers
var borrow,i : Byte;
begin
 borrow := 0;
 for i := 0 to 7 do
  begin
   if v2[i] < v1[i]+borrow then
    begin
     result[i] := 256 + v2[i] - v1[i] - borrow;
     borrow := 1;
    end
    else begin
          result[i] := v2[i] - v1[i] - borrow;
          borrow := 0;
         end;
  end;
end;

function I64toStr(I64 : TI64) : string;
//convert a 64 bit integer to a string
const d = 10;
      hb = $80;
      msk = $7f;
var i,j,b,c : byte;
    sum : word;
begin
 result := '';
 repeat
   I64[8] := 0;
   for i := 1 to 64 do
    begin
     for j := 8 downto 1 do
      begin
       c := I64[j-1] shr 7;
       I64[j] := ((I64[j] and msk) shl 1) or c;
      end;//for j
      if I64[8] >= d then begin
                           b := 1;
                           I64[8] := I64[8] - d;
                           end
      else b := 0;
     I64[0] := ((I64[0] and msk) shl 1) or b;
    end;//for i
   insert(chr(ord('0') + I64[8]),result,0);
  sum := 0;
  for j := 0 to 7 do inc(sum,I64[j]);
 until sum = 0;
end;

function I64tofloat(I64 : TI64) : extended;
var i : byte;
    w : extended;
begin
 result := 0;
 w := 1;
 for i := 0 to 7 do
  begin
   result := result + I64[i]*w;
   w := w * 256;
  end;
end;

end.

